<?php
declare(strict_types=1);

namespace Edara\Includes;

use EdaraAPI\EdaraAPI;
use WP_User;

class CustomerService extends AbstractService
{
    /**
     * Find customer by email
     *
     * @param string $email
     * @return int
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function findCustomerByEmail(string $email): int
    {
        $customer = $this->api()->findCustomerIdByEmail($email);
        $customerId = array_get($customer, 'result.id', 0);

        return $customerId;
    }

    /**
     * Create external customer
     *
     * @param \WP_User $user
     * @return int
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    protected function createExternalCustomer(WP_User $user): int
    {
        $data = [
            'name' => $user->nickname ?: $user->user_email,
            'email' => $user->user_email,
            'mobile' => $user->phone,
            'pricing_type' => 'EndUser',
            'payment_type' => 'Credit',
        ];
        $output = $this->api()->createCustomer($data);

        if ("409" == array_get($output, 'status_code')) {
            $data['mobile'] = '+2' . $data['mobile'];
            $output = $this->api()->createCustomer($data);
        }

        // @TODO Check and log error if not result
        return (int) array_get($output, 'result', 0);
    }

    /**
     * Find external customer id or create a new one
     *
     * @param int $customerId
     * @return int
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function findExternalCustomerIdOrCreateNew(int $customerId, array $billingInfo):int
    {
        ## Find external customer id by billing info
        if (!$customerId) {
            return (int)$this->findOrCreateExternalCustomerByBillingInfo($billingInfo);
        }

        $externalId = get_user_meta($customerId, 'external_id', true);
        #1 Simply find the external customer id associated with user
        if ($externalId) {
            return (int)$externalId;
        }
        $userDetails = get_userdata($customerId);

        #2 Try to find customer Id by email
        if ($userDetails->user_email) {
            $externalId = $this->findCustomerByEmail($userDetails->user_email);
            if ($externalId) {
                return (int)$externalId;
            }
        }

        #3 Create a new customer and save the external customer id in user
        $externalId = $this->createExternalCustomer($userDetails);
        update_user_meta($customerId, 'external_id', $externalId);

        return (int)$externalId;
    }

    /**
     * Find or create an external customer
     *
     * @param array $billingInfo
     * @return int
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    private function findOrCreateExternalCustomerByBillingInfo(array $billingInfo):int
    {
        #1 Find customer by email if exist
        $customerExternalId = (int) $this->findCustomerByEmail(array_get($billingInfo, 'email'));

        if ($customerExternalId) {
            return (int)$customerExternalId;
        }

        #2 Prepare user object
        $userDetails = new WP_User();
        $firstName = array_get($billingInfo, 'first_name');
        $lastName = array_get($billingInfo, 'last_name');

        $userDetails->nickname = "$firstName $lastName" ;
        $userDetails->first_name = $firstName;
        $userDetails->last_name = $lastName;
        $userDetails->user_email = array_get($billingInfo, 'email');
        $userDetails->phone = array_get($billingInfo, 'mobile');

        #3 Create a new customer on Edara
        return (int)$this->createExternalCustomer($userDetails);
    }
}
