<?php
declare(strict_types=1);
  
namespace Edara\Includes;

class StockService extends AbstractService
{
    /**
     * Get the stock item balance
     *
     * @param int $externalProductId
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function retrieveStock(int $externalProductId): array
    {
        $response = $this->api()->retrieveStockBalance($externalProductId);
        $result = array_get($response, 'result', []);
        $result = array_filter($result, fn($item) => (EDARA_MAIN_WAREHOUSE_ID == $item['warehouse_id']) ? $item : null);

        $reservedBalance = array_sum(array_column($result,'reserved_balance'));
        $balance = array_sum(array_column($result,'balance'));

        return ['balance' => $balance, 'reserved_balance' => $reservedBalance];
    }

    /**
     * Retrieve global stock balance
     *
     * @param int $externalProductId
     * @return float
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function retrieveGlobalStockBalance(int $externalProductId):float
    {
        $response = $this->api()->retrieveStockGlobalBalance($externalProductId);

        return array_get($response, 'result.global_balance');
    }

    /**
     * Get stock item balance
     *
     * @param int $externalProductId
     * @return float
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function retrieveStockBalance(int $externalProductId):float
    {
        return array_get($this->retrieveStock($externalProductId), 'balance');
    }

    /**
     * Retrieve the available stock balance
     *
     * @param int $externalProductId
     * @return float
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function retrieveStockAvailableBalance(int $externalProductId): float
    {
        $response = $this->retrieveStock($externalProductId);
        $balance = (intval(array_get($response, 'balance')) - intval(array_get($response, 'reserved_balance')));

        return (float) $balance;
    }

    /**
     * Get stock item reserved balance
     *
     * @param int $externalProductId
     * @return float
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function retrieveStockReservedBalance(int $externalProductId):float
    {
        return array_get($this->retrieveStock($externalProductId), 'reserved_balance');
    }

    /**
     * Update product stock balance
     *
     * @param array $data
     */
    public function updateProductStockBalance(array $data):void
    {
        $externalProductId = array_get($data, 'data.id');
        $localBalance = intval(array_get($data, 'message_attributes.LocalBalance'));
        $reservedBalance = intval(array_get($data, 'message_attributes.LocalReservedBalance'));
        $netBalance = intval($localBalance - $reservedBalance);
        $product = (new ProductService())->findProductByExternalId($externalProductId);

        if ($product) {
            $product->set_stock_quantity($netBalance);
            $product->save();

            $data = ['id' => $product->get_id(), 'external_id' => $externalProductId, 'new_balance' => $product->get_stock_quantity(),];
            $myFileLink2 = fopen((ABSPATH . 'stock_balance_change'), 'a+');
            fwrite($myFileLink2, json_encode($data) . "\n\n");
            fclose($myFileLink2);
        }
    }
}
