<?php
declare(strict_types=1);

namespace EdaraAPI;

use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Psr7\Response;

class EdaraAPI
{

    /**
     * @var \GuzzleHttp\Client
     */
    protected $client;

    /**
     * Hold the class instance
     * @var \EdaraAPI\EdaraAPI
     */
    private static $instance;

    /**
     * The API key
     * @var string
     */
    protected $apiKey;

    private $url = [
        'update_product' => '/v2.0/stockItems',
        'create_product' => '/v2.0/stockItems',
        'listing_products' => '/v2.0/stockItems',
        'product_details' => '/v2.0/stockItems/%s',
        'delete_product' => '/v2.0/stockItems/DeleteByCode/%s',
        'find_customer' => '/v2.0/customers/%s',
        'find_customer_by_email' => '/v2.0/customers/FindByEmail/%s',
        'find_customers' => '/v2.0/customers',
        'delete_customer' => '/v2.0/customers/%s',
        'create_customer' => '/v2.0/customers',
        'list_orders' => '/v2.0/salesOrders',
        'create_order' => '/v2.0/salesOrders',
        'update_order' => '/v2.0/salesOrders',
        'delete_order' => '/v2.0/salesOrders/%s',
        'cancel_order' => '/v2.0/salesOrders/CancelByCode/%s',
        'update_order_status' => '/v2.0/salesOrders/UpdateSalesOrderStatus/%s',
        'retrieve_order_by_code' => '/v2.0/salesOrders/FindByCode/%s',
        'retrieve_stock_balance' => '/v2.0/stockItems/%s/balance',
        'retrieve_stock_global_balance' => '/v2.0/stockItems/GlobalBalance/%s',
        'create_classification' => '/v2.0/stockItems/classifications',
        'retrieve_classification' => '/v2.0/stockItems/classifications',
        'create_sales_return' => '/v2.0/salesreturns',
        'list_bundles' => '/v2.0/salesBundles',
    ];

    // to prevent initiation with outer code.
    private function __construct(string $apiKey)
    {
        $this->apiKey = $apiKey;
        $this->client = $this->initiateNewClient($this->apiKey, 'https://api.edara.io/v2.0/');
    }

    /**
     * @return \EdaraAPI\EdaraAPI
     */
    public static function instance(): EdaraAPI
    {
        if (!self::$instance) {
            self::$instance = new EdaraAPI(EDARA_BEARER_TOKEN);
        }

        return self::$instance;
    }

    /**
     * List all products
     *
     * @param int $limit
     * @param int $offset
     * @return false|string
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function listStockItems(int $limit = 10, int $offset = 0):array
    {
        $params = ['limit' => $limit, 'offset' => $offset];
        $url = $this->url['listing_products'];

        return $this->get($url, $params);
    }

    public function listProducts(int $limit = 10, int $offset = 0):array
    {
        return $this->listStockItems(...func_get_args());
    }

    /**
     * Delete product by Id
     *
     * @param int $stockItemId
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function deleteStockItemById(int $stockItemId):array
    {
        return $this->delete(sprintf($this->url['product_details'], $stockItemId));
    }

    /**
     * Get stock item
     *
     * @param $stockItemId
     * @return mixed
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function findStockItem(string $stockItemId):array
    {
        $url = sprintf($this->url['product_details'], $stockItemId);

        return $this->get($url);
    }

    /**
     * find customer by id
     *
     * @param $customerId
     * @return mixed
     */
    public function findCustomerById(string $customerId):array
    {
        $url = sprintf($this->url['find_customer'], $customerId);

        return $this->get($url);
    }

    /**
     * Find customer by Email
     *
     * @param string $email
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function findCustomerIdByEmail(string $email)
    {
      if (false) {
        $url = sprintf($this->url['find_customer_by_email'], trim($email));

        return $this->get($url);
      }
    }

    /**
     * Delete customer by Id
     *
     * @param $customerId
     * @return mixed
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function deleteCustomerById(string $customerId):array
    {
        $url = sprintf($this->url['delete_customer'], $customerId);

        return $this->delete($url);
    }

    /**
     * Create a new customer
     *
     * @param $data
     * @return mixed
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function createCustomer(array $data):array
    {
        return $this->post($this->url['create_customer'], [], $data);
    }

    /**
     * Find list of customers
     *
     * @param int $limit
     * @param int $offset
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function findCustomers(int $limit = 10, int $offset = 1):array
    {
        return $this->get($this->url['find_customers'], [
            'limit' => $limit,
            'offset' => $offset,
        ]);
    }

    /**
     * Create classification
     *
     * @param array $data
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function createClassification(array $data):array
    {
        return $this->post($this->url['create_classification'], [], $data);
    }

    /**
     * Find classification
     *
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function retrieveClassifications():array
    {
        return $this->get($this->url['retrieve_classification']);
    }

    /**
     * Find classification by Id
     *
     * @param int $classificationId
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function findClassificationsById(int $classificationId):array
    {
        $response = $this->get($this->url['retrieve_classification']);
        $result = array_get($response, 'result', []);
        // phpcs:ignore
        $classification = array_filter($result, fn($classification) => array_get($classification, 'classification_id') === $classificationId);
        if ($classification) {
            return reset($classification);
        }

        return [];
    }

    /**
     * Create stock item
     *
     * @param array $data
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function createStockItem(array $data = []):array
    {
        return $this->post($this->url['create_product'], [], $data);
    }

    /**
     * Update stock item
     *
     * @param array $data
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function updateStockItem(array $data = []): array
    {
        $url = $this->url['update_product'];

        return $this->put($url, [], $data);
    }

    /**
     * Create sales return
     *
     * @param array $data
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function createSalesReturn(array $data = []): array
    {
        return $this->post($this->url['create_sales_return'], [], $data);
    }

    /**
     * Create a new order
     *
     * @param $data
     * @return mixed
     */
    public function createOrder(array $data = []):array
    {
        return $this->post($this->url['create_order'], [], $data);
    }

    /**
     * Update order
     *
     * @param $data
     * @return mixed
     */
    public function updateOrder(array $data):array
    {
        $url = $this->url['update_order'];

        return $this->put($url, [], $data);
    }

    /**
     * Retrieve order details by code
     *
     * @param $orderCode
     * @return array|mixed
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function retrieveOrderByCode(string $orderCode):array
    {
        $url = sprintf($this->url['retrieve_order_by_code'], $orderCode);

        return $this->get($url);
    }

    /**
     * Find orders
     *
     * @param int $limit
     * @param int $offset
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function findOrders(int $limit, int $offset): array
    {
        return $this->get($this->url['list_orders'], [
            'limit' => $limit,
            'offset' => $offset,
        ]);
    }

    /**
     * Delete order
     *
     * @param int $orderId
     * @return mixed
     */
    public function deleteOrder(int $orderId):array
    {
        $url = sprintf($this->url['delete_order'], $orderId);

        return $this->delete($url);
    }

    /**
     * Update order status
     *
     * @param string $orderCode
     * @param string $status
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function updateOrderStatus(string $orderCode, string $status):array
    {
        $url = sprintf($this->url['update_order_status'], $orderCode);

        return $this->put($url, [], ['order_status' => $status]);
    }

    /**
     * Cancel order
     *
     * @param string $orderCode
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function cancelOrder(string $orderCode): array
    {
        $url = sprintf($this->url['cancel_order'], $orderCode);

        return $this->put($url);
    }

    /**
     * Retrieve stock balance of stock item
     *
     * @param int $stockItemId
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function retrieveStockBalance(int $stockItemId):array
    {
        $url = sprintf($this->url['retrieve_stock_balance'], $stockItemId);

        return $this->get($url);
    }

    /**
     * Retrieve stock global balance
     *
     * @param int $stockItemId
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function retrieveStockGlobalBalance(int $stockItemId):array
    {
        $url = sprintf($this->url['retrieve_stock_global_balance'], $stockItemId);

        return $this->get($url);
    }

    /**
     * List all bundles
     *
     * @param int $limit
     * @param int $offset
     * @return false|string
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function listBundles(int $limit = 10, int $offset = 0): array
    {
        $params = ['limit' => $limit, 'offset' => $offset];

        return $this->get($this->url['list_bundles'], $params);
    }

    /**
     * Delete request
     *
     * @param $url
     * @return mixed
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    protected function delete(string $url):array
    {
        return $this->parseResponse($this->client->delete($url));
    }

    /**
     * Put request
     *
     * @param string $url
     * @param array $formParams
     * @param array $body
     * @return mixed
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    protected function put(string $url, array $formParams = [], array $body = []):array
    {
        return $this->parseResponse($this->client->put($url, [
            'form_params' => $formParams,
            'json' => $body,
        ]));
    }

    /**
     * Post request
     *
     * @param string $url
     * @param array $formParams
     * @param array $body
     * @return mixed
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    protected function post(string $url, array $formParams, array $body = []):array
    {
        $message = "Post request: " . date("Y-m-d H:i:s") . PHP_EOL;
        file_put_contents("log_requests.log",print_r($message,true),FILE_APPEND);
        file_put_contents("log_requests.log","URL: " . $url,FILE_APPEND);
        file_put_contents("log_requests.log","Form params: " . print_r($formParams,true),FILE_APPEND);
        file_put_contents("log_requests.log","Body: " . print_r($body,true),FILE_APPEND);
        return $this->parseResponse($this->client->post($url, [
            'form_params' => $formParams,
            'json' => $body,
        ]));
    }

    /**
     * Get request
     *
     * @param string $url
     * @param array $params
     * @return mixed
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    private function get(string $url, array $params = []):array
    {
        $message = "Get request: " . date("Y-m-d H:i:s") . PHP_EOL;
        file_put_contents("log_requests.log",print_r($message,true),FILE_APPEND);
        file_put_contents("log_requests.log","URL: " . $url,FILE_APPEND);
        file_put_contents("log_requests.log","params: " . print_r($params,true),FILE_APPEND);
        return $this->parseResponse($this->client->get($url, ['query' => $params]));
    }

    /**
     * Parse the response from Edara API
     *
     * @param \GuzzleHttp\Psr7\Response $response
     * @return mixed
     */
    protected function parseResponse(Response $response):array
    {
        $message = "Request response: " . date("Y-m-d H:i:s") . PHP_EOL;
        file_put_contents(ABSPATH . "log_requests.log",print_r($message,true),FILE_APPEND);
        file_put_contents(ABSPATH . "log_requests.log",print_r($response,true),FILE_APPEND);
        if (200 === $response->getStatusCode()) {
            return json_decode($response->getBody()->getContents(), true);
        }

        throw new ClientException('Invalid http client request.');
    }

    /**
     * Initiate a new client
     *
     * @param $apiKey
     * @param $baseUrl
     * @return \GuzzleHttp\Client
     */
    private function initiateNewClient(string $apiKey, string $baseUrl):\GuzzleHttp\Client
    {
        return new \GuzzleHttp\Client(
            [
                'base_uri' => $baseUrl,
                'headers' => [
                    'Authorization' => 'Bearer ' . $apiKey,
                    'Accept' => 'application/json',
                    'Content-Type' => 'application/json',
                ],
            ]
        );
    }
}
